;; A common security method used for online banking is to ask the user
;; for three random characters from a passcode. For example, if the
;; passcode was 531278, they may asked for the 2nd, 3rd, and 5th
;; characters; the expected reply would be: 317.

;; The text file, keylog.txt, contains fifty successful login
;; attempts.

;; Given that the three characters are always asked for in order,
;; analyse the file so as to determine the shortest possible secret
;; passcode of unknown length.


(defvar *the-file* "/home/mortis/personal/projects/project-euler.net/data-files/keylog.txt")

(defvar *logins* nil)

(setf *logins* nil)

(with-open-file (in *the-file* :direction :input)
  (loop for line = (read-line in nil nil) then (read-line in nil nil)
	while line
	do
	(push line *logins*)))

;; (require :cl-ppcre)
(defun grep (pat things)
  (loop for thing in things
	when (cl-ppcre:scan-to-strings pat thing)
	collect thing))

(defun explode-string (str)
  (loop for ch across str
	collect (format nil "~a" ch)))

(defun dig-string->nums (str)
  (loop for dig in (explode-string str)
	collect (read-from-string dig)))

;; (dig-string->nums "123")

;; (grep "31" *logins*)
;; => ("319" "316" "319" "319" "318" "316" "731" "318" "319")
;; this tells us the following: 
;;   1 always comes after 3
;;   (9, 6, and 8) come after 1

(defun ht-pushval (ht key val)
  (symbol-macrolet ((ht-get (gethash key ht)))
      (multiple-value-bind (lst existed)
	  ht-get
	(if existed
	    (setf ht-get
		  (cons val lst))
	    (setf ht-get
		  (list val))))))

(defun make-constraint-table ()
  (let ((ht (make-hash-table :test #'equal)))
    (loop 
     for login in *logins*
     for nums = (dig-string->nums login) then (dig-string->nums login)
     for frst = (first nums)  then (first nums)
     for scnd = (second nums) then (second nums)
     for thrd = (third nums)  then (third nums)
     do
     (format t "idx:~a < idx:~a~&"
	     frst scnd)
     (format t "idx:~a < idx:~a~&"
	     scnd thrd)
     (ht-pushval ht frst scnd)
     (ht-pushval ht scnd thrd)
     (if (not (gethash thrd ht))
	 (setf (gethash thrd ht) nil)))
    ht))

(defun list-unique (things)
  (let ((ht (make-hash-table :test #'equal))
	(result (list)))
    (loop for thing in things
	  do
	  (when (not (gethash thing ht))
	    (setf (gethash thing ht) t)
	    (push thing result)))
    (reverse result)))

(let ((ht (make-constraint-table)))
  (loop for key being the hash-keys of ht
	do
	(format t "~a: ~a~&"
		key
		(list-unique (gethash key ht)))))





;; 7: (6 1 2 3 9)
;; 1: (9 8 2 0 6)
;; 2: (9 0 8)
;; 3: (1 6 8)
;; 8: (0 9)
;; 6: (8 9 2 0)
;; 9: (0)

;; 73162890
;; (length "73162890") => 8

(defun index (thing str)
  (loop for ii = 0 then (incf ii)
	for ch across str
	do
	(when (equal ch thing)
	  (return-from index ii)))
  nil)

;; (index #\7 "137")

(defun string->chars (str)
  (loop for ch across str
	collect ch))

;; (string->chars "fo")

(loop 
 for login in *logins*
 for chrs = (string->chars login)
 for frst = (first chrs) then (first chrs)
 for scnd = (second chrs) then (second chrs)
 for thrd = (third chrs) then (third chrs)
 for frstidx = (index frst "73162890") then (index frst "73162890")
 for scndidx = (index scnd "73162890") then (index scnd "73162890")
 for thrdidx = (index thrd "73162890") then (index thrd "73162890")
 for is-sorted = (and (< frstidx scndidx) (< scndidx thrdidx))
 then (and (< frstidx scndidx) (< scndidx thrdidx))
 do
 (format t "~a indexes: ~a ~a,~a,~a~&"
	 login is-sorted frstidx scndidx thrdidx))

      